/*
** Module   : ControlEntity.java
** Abstract : server side control widget parent class
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23486 Created initial version.
** 002 GES 20060114   @23908 Removed totally useless screen-value attr.
** 003 SIY 20060202   @24235 Added support for TO option.
** 004 SIY 20060203   @24247 Added force-label flag.
** 005 GES 20060204   @24262 Added getType() method to obtain the data
**                           type the widget represents.  Also converted
**                           the instance data to primitives instead of
**                           wrappers.
** 006 SIY 20060209   @24435 Implemented tracking of widget data type.
**                           Fixed method name. Methods from derived 
**                           classes moved here.
** 007 SIY 20060426   @25913 Changed signature of getNextTabItem.
** 008 SIY 20060515   @26150 Moved COLUMN-LABEL attribute from 
**                           GenericWidget.
** 009 SIY 20060720   @28148 Added support for INDEX attribute.
** 010 GES 20060803   @28370 Some cleanup and refactoring.
** 011 NVS 20060918   @29643 Added setAttrSpace() method.
** 012 SIY 20061102   @30936 Added alignment mode constants and method.
** 013 NVS 20070426   @33243 For the first time, implemented 
**                           getNextTabItem() method, which was a stub.
** 014 GES 20080320   @37596 Changed getType() to getDataClass() to avoid
**                           method name conflicts with the TYPE attr.
** 015 SIY 20090604   @42611 Forcing labels resets NO-LABEL flag.
** 016 SIY 20101122          Added generic in one method return type in order
**                           to make it compatible with other changes.
** 017 SIY 20111101          Updated imports.
** 018 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 019 CA  20131115          Changes related to properly instantiate dynamic resources.
** 020 HC  20140911          GUI coordinate support - data types refactoring.
** 021 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 022 HC  20141207          Implemented runtime support for widget coordinate attributes.
** 023 GES 20141124          Added modified attr.  Moved setColon() to BaseEntity since customer
**                           code shows it can be used widely for other widgets (undocumented
**                           feature).
** 024 GES 20150104          Modifications for more correct modified attribute support.
** 025 CA  20150105          Changes for SIDE-LABEL-HANDLE runtime support.
** 026 CA  20150123          Fixed handling of side-labels in shared frames.
** 027 CA  20150219          Added support for DATA-TYPE attribute.
** 028 IAS 20150220          added setAutoresize(boolean).
*/ 

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;
import com.goldencode.p2j.ui.client.format.*;

/**
 * Represents widget functionality that is associated with all widgets that
 * represent data OR which provide a control feature (like a button).
 */
public class ControlEntity<T extends ControlConfig>
extends BaseEntity<T>
{
   /** Alignment mode - align widget left */
   public static final int ALIGN_LEFT  = 1;

   /** Alignment mode - align widget right */
   public static final int ALIGN_RIGHT = 2;
   
   /** Alignment mode - align widget to label colon */
   public static final int ALIGN_COLON = 3;
   
   /** Type of the variable in the widget. */
   private Class<?> dataTypeClass = null;
   
   /**
    * Reference to the side-label widget (as returned by the SIDE-LABEL-HANDLE).  When
    * <code>null</code>, it means no side-label exists (thus SIDE-LABEL-HANDLE is unknown).
    */
   private LiteralWidget sideLabel = null;
   
   /**
    * Constructor for use by sub-classes.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    ctrl
    *           Configuration container.
    */
   public ControlEntity(boolean dynamic, T ctrl)
   {
      super(dynamic, ctrl);
   }
   
   /**
    * Accesses the data type which the widget represents. 
    *
    * @return   The class of the value that the widget represents or
    *           <code>null</code> if this widget does not represent any
    *           data (e.g. the widget is a button).
    */
   public Class<?> getDataClass()
   {
      if (dataTypeClass == null)
      {
         BaseDataType var = DisplayFormat.instanceOfType(config.dataType);
         
         if (var == null)
         {
            throw new RuntimeException("Attempt to set invalid data type " +  config.dataType);
         }
         
         dataTypeClass = var.getClass();
      }

      return dataTypeClass;
   }

   /**
    * Sets the column label.
    * 
    * @param   clabel
    *          column label text
    */
   public void setColumnLabel(character clabel)
   {
      setColumnLabel(clabel.isUnknown() ? "" : clabel.toStringMessage());
   }

   /**
    * Sets the column label.
    * 
    * @param   clabel
    *          column label text
    */
   public void setColumnLabel(String clabel)
   {
      config.clabel = clabel;
      pushScreenDefinition();
   }
   
   /**
    * Force displaying of label.
    *
    * @param    force
    *           New value for the flag.
    */
   public void setForceLabel(logical force)
   {
      setForceLabel(force.booleanValue());
   }

   /**
    * Force displaying of label.
    *
    * @param    force
    *           New value for the flag.
    */
   public void setForceLabel(boolean force)
   {
      if (force)
         config.labels = true;
      
      config.forceLabel = force;
      pushScreenDefinition();
   }
   
   /**
    * Getter for the NEXT-TAB-ITEM read-only attribute.
    *
    * @return   The value of the NEXT-TAB-ITEM attribute.
    */
   public GenericWidget<?> getNextTabItem()
   {
      return LogicalTerminal.nextTabItem(getFrame(), this);
   }

   /**
    * Set alignment.
    * 
    * @param   mode
    *          Set alignment mode.
    */
   public void setAlign(int mode)
   {
      config.align = mode;
      pushScreenDefinition();
   }
   
   /**
    * Sets the rightmost column of the widget.
    * 
    * @param   pos
    *          The 1-based column number. This is converted to a 0-based
    *          column number for the internal runtime.
    */
   public void setTo(NumberType pos)
   {
      setTo(pos.intValue());
   }

   /**
    * Sets the rightmost column of the widget.
    * 
    * @param   pos
    *          The 1-based column number. This is converted to a 0-based
    *          column number for the internal runtime.
    */
   public void setTo(double pos)
   {
      double newCol = pos - 1;
      config.setTo(newCol);
      pushScreenDefinition();
   }
   
   /**
    * Gets the column number to which colon between label and widget should be
    * aligned. Note that Progress does not have corresponding attribute so
    * calls to this method are unlikely.
    * 
    * @return  The 1-based column of the desired colon position. This is
    *          converted from a 0-based column number for the internal runtime.
    */
   public decimal getColon()
   {
      double col = config.getColon() + 1;
      return new decimal(col);
   }

   /**
    * Sets new value of DATA-TYPE attribute.
    *
    * @param dataType
    *        new value of DATA-TYPE attribute
    */
   @Override
   public void setDataType(character dataType)
   {
      if (dataType == null || dataType.isUnknown())
      {
         String txt = 
           "**Attribute DATA-TYPE for the FILL-IN widget has an invalid value of UNKNOWN. (4056)";
         ErrorManager.recordOrShowError(4056, txt, false, false, false);
      }
      
      setDataType(dataType.toStringMessage());
   }
   
   /**
    * Set type of the bound variable.
    * 
    * @param   dataType
    *          New data type name.
    */
   @Override
   public void setDataType(String dataType)
   {
      config.dataType = dataType;
      
      // resolve the class now.
      getDataClass();

      pushScreenDefinition();
   }

   /**
    * Queries the DATA-TYPE attribute of this widget.
    *
    * @return A character value that represents the data type of the field
    *         associated with the widget.
    */
   @Override
   public character getDataType()
   {
      return new character(config.dataType);
   }

   /**
    * Queries if the SCREEN-VALUE attribute for this widget was changed.
    * 
    * @return  If <code>true</code>, the widget's screen value was changed.
    */
   public logical isModified()
   {
      // TODO: 4GL docs suggest we need to realize the widget when this is called
      //       and thorough compatibility testing is really needed 
      UnimplementedFeature.todo("MODIFIED attribute behavior needs more testing");
      return new logical(config.modified);
   }
   
   /**
    * Defines if the SCREEN-VALUE attribute for this widget has changed.
    * 
    * @param   modified
    *          <code>true</code> to mark the screen value as changed.
    */
   public void setModified(boolean modified)
   {
      UnimplementedFeature.todo("MODIFIED attribute behavior needs more testing");
      config.setModified(modified);
      pushScreenDefinition();
   }
   
   /**
    * Defines if the SCREEN-VALUE attribute for this widget has changed.
    * 
    * @param   modified
    *          <code>true</code> to mark the screen value as changed.
    */
   public void setModified(logical modified)
   {
      if (modified == null || modified.isUnknown())
      {
         UnimplementedFeature.todo("MODIFIED attribute unknown behavior is not known.");
         return;
      }
      
      setModified(modified.booleanValue()); 
   }
   
   /**
    * Set new value to NO-LABELS flag.
    * 
    * @param   noLabels
    *          New value of the flag.
    */
   public void setNoLabels(boolean noLabels)
   {
      config.labels = !noLabels;
      pushScreenDefinition();
   }
   
   /**
    * Sets the value of the ATTR-SPACE attribute.
    * 
    * @param   attrSpace
    *          The new boolean value for the ATTR-SPACE attribute.
    */
   public void setAttrSpace(boolean attrSpace)
   {
      config.attrSpace = attrSpace;
   }

   /**
    * Blocks attempts to get the TITLE attribute as it is undefined.
    * 
    * @return Always throws RuntimeException. 
    */
   public character getTitle() 
   {
      throw new RuntimeException("TITLE attribute is undefined");
   }

   /**
    * Blocks attempts to set the TITLE attribute as it is undefined.
    * 
    * @param    t
    *           Unused.
    */
   public void setTitle(character t) 
   {
      throw new RuntimeException("TITLE attribute is undefined");
   }

   /**
    * Blocks attempts to set the TITLE attribute as it is undefined.
    * 
    * @param    t
    *           Unused.
    */
   public void setTitle(String t)
   {
      throw new RuntimeException("TITLE attribute is undefined");
   }

   /**
    * Set INDEX attribute value.
    * 
    * @param    index
    *           New value for the attribute.
    */
   public void setIndex(int index)
   {
      this.config.index = index;
   }

   /**
    * Set INDEX attribute value.
    * 
    * @param    index
    *           New value for the attribute.
    */
   public void setIndex(NumberType index)
   {
      if (index.isUnknown())
         return;
      
      setIndex(index.intValue());
   }

   /**
    * Get current value of the INDEX attribute.
    * 
    * @return   The current value of the attribute.
    */
   public integer getIndex()
   {
      return new integer(config.index);
   }
   
   /**
    * Get current value of the AUTO-RESIZE attribute.
    * 
    * @return   The current value of the attribute.
    */
   public logical isAutoResize()
   {
      return new logical(config.autoResize);
   }
   
   /**
    * Set AUTO-RESIZE attribute value.
    * 
    * @param    autoResize
    *           New value for the attribute.
    */
   public void setAutoResize(logical autoResize)
   {
      if (autoResize.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, String.format(
               "Unable to assign UNKNOWN value to attribute AUTO-RESIZE on %s %s.", type(),
                  frame.getName(this)), false);
      }
      setAutoResize(autoResize.booleanValue());
   }

   /**
    * Set AUTO-RESIZE attribute value.
    * 
    * @param    autoResize
    *           New value for the attribute.
    */
   public void setAutoResize(boolean autoResize)
   {
      config.autoResize = autoResize;
   }

   /**
    * Obtains the text widget that is used as the side label for this widget. This is the
    * equivalent of the SIDE-LABEL-HANDLE attribute of COMBO-BOX, EDITOR, FILL-IN, RADIO-SET,
    * SELECTION-LIST, SLIDER and TEXT widgets.
    *
    * @return   The handle to the text widget that is the side label of this widget.
    */
   public handle getSideLabelHandle()
   {
      return sideLabel == null ? new handle() : new handle(sideLabel);
   }

   /**
    * The method that actually performs assignment of the HEIGHT-CHARS attribute.  
    * 
    * @param    height
    *           The attribute value.
    */
   @Override
   protected void setHeightCharsWorker(double height)
   {
      config.autoResize = false;
      super.setHeightCharsWorker(height);
   }

   /**
    * The method that actually performs assignment of the WIDTH-CHARS attribute.  
    * 
    * @param    width
    *           The attribute value.
    */
   @Override
   protected void setWidthCharsWorker(double width)
   {
      config.autoResize = false;
      super.setWidthCharsWorker(width);
   }

   /**
    * The method that actually performs assignment of the WIDTH=PIXELS attribute.  
    * 
    * @param    widthPixels
    *           The attribute value.
    */
   @Override
   protected void setWidthPixelsWorker(int widthPixels)
   {
      config.autoResize = false;
      super.setWidthPixelsWorker(widthPixels);
   }

   /**
    * The method that actually performs assignment of the HEIGHT-PIXELS attribute.  
    * 
    * @param    heightPixels
    *           The attribute value.
    */
   @Override
   protected void setHeightPixelsWorker(int heightPixels)
   {
      config.autoResize = false;
      super.setHeightPixelsWorker(heightPixels);
   }
   
   /**
    * Check if a side-label can be attached.
    * <p>
    * This is possible only if the widget accepts a side-label the
    * ({@link ControlConfig#possibleSideLabel} flag is <code>true</code>) and also the frame is
    * configured with the SIDE-LABELS option.
    * 
    * @return   <code>true</code> if the SIDE-LABEL-HANDLE needs to return a valid handle.
    */
   protected boolean hasSideLabelHandle()
   {
      FrameConfig fconfig = frame.getFrameWidget().config;
      
      // options which affect this:
      // 1. frame
      // SIDE-LABELS
      // NO-LABELS
      // 2. widget
      // explicit label (forceLabel flag)
      // no label (label field to null)
      // NO-LABELS
      
      return config.possibleSideLabel && 
             fconfig.sideLabels       &&
             ((fconfig.labels && config.labels) ||
              (!fconfig.labels && config.labels && config.forceLabel)); 
   }
   
   /**
    * Process the side-label for this widget.
    * <p>
    * If the widget accepts a side-label, it will construct it if {@link #sideLabel} is not 
    * already set.
    * 
    * @return   The ID for the side-label widget (as reported by {@link ControlConfig#sideLabelId})
    *           or <code>-1</code> if none exists.
    */
   @Override
   protected int processSideLabel()
   {
      if (!frame.isInsideSetup() && hasSideLabelHandle() && config.label != null)
      {
         // do this only once. nothing is initialized here.  
         // all info will be coming from the client-side

         if (sideLabel == null)
         {
            if (config.sideLabelId == -1)
            {
               sideLabel = new LiteralWidget();
               sideLabel.setFrame(this.getFrame());
   
               config.sideLabelId = WidgetId.nextID().getId();
               sideLabel.config.id = new WidgetId(config.sideLabelId);
   
               LogicalTerminal.registerWidget(sideLabel, config.sideLabelId);
               ConfigManager.getInstance().addWidgetConfig(sideLabel.config);
            }
            else
            {
               // someone else has created it (i.e. a shared frame), retrieve it from the registry
               sideLabel = (LiteralWidget) LogicalTerminal.getWidgetForId(config.sideLabelId);
            }
         }
         
         frame.getFrameBufferRef().putScreenValue(config.sideLabelId, new character(config.label));
      }
      
      return config.sideLabelId;
   }
   
   /**
    * Set this widget's config directly, to share it with its associated widget from the master
    * frame.
    * <p>
    * The widget's {@link #config} instance will be replaced with the active config specified by
    * the {@link WidgetConfig#id config.id}.
    * <p>
    * Must be called only when the shared frame is being imported.
    */
   @Override
   protected void setSharedConfig()
   {
      super.setSharedConfig();
      
      // if there is a side-label, register it too
      if (sideLabel != null)
      {
         sideLabel.setSharedConfig();
      }
   }
   
   /**
    * Attempt to remove the shared config from the config manager.
    * <p>
    * Must be called only when the shared frame is getting out of scope.
    */
   @Override
   protected void removeSharedConfig()
   {
      super.removeSharedConfig();
      
      if (sideLabel != null)
      {
         sideLabel.removeSharedConfig();
      }
   }
}
